# fmt: off
from collections import defaultdict

import numpy as np

import sco_py.expr as sco_osqp_expr
import sco_py.sco_osqp.osqp_utils as osqp_utils
from sco_py.sco_osqp.osqp_utils import (OSQPLinearConstraint, OSQPLinearObj,
                                 OSQPQuadraticObj, OSQPVar)

# fmt: on


class Prob(object):
    """
    Sequential convex programming problem with a scalar objective. A solution is
    found using the l1 penalty method.
    """

    def __init__(self, callback=None):
        """
        _vars: variables in this problem
        _osqp_vars: a set of all the osqp_vars in this problem. This will be used to
            construct the x vector whenever a call is made to the OSQP solver.

        _quad_obj_exprs: list of quadratic expressions in the objective
        _nonquad_obj_exprs: list of non-quadratic expressions in the objective
        _approx_obj_exprs: list of approximation of the non-quadratic
            expressions in the objective

        _nonlin_cnt_exprs: list of non-linear constraint expressions
        _penalty_exprs: list of penalty term expressions (approximations of the
            non-linear constraint expressions in _nonlin_cnt_exprs)

        _osqp_quad_objs: list of OSQPQuadraticObjs that keep track of the quadratic objectives
            that will be passed to the QP
        _osqp_lin_objs: list of OSQPLinearObjectives that keep track of the linear objectives
            that will be passed to the QP
        _osqp_lin_cnt_exprs: list of OSQPLinearConstraints that keep track of the linear constraints
            that will be passed to the QP

        _osqp_penalty_cnts: list of OSQP constraints that are generated when
            adding the hinge and absolute value terms from the penalty terms.

        _bexpr_to_osqp_expr: dictionary that caches quadratic bound expressions
            with their corresponding Gurobi expression
        """
        self._vars = set()
        self._osqp_vars = set()
        if callback is not None:
            self._callback = callback
        else:

            def do_nothing():
                pass

            self._callback = do_nothing

        self._quad_obj_exprs = []
        self._nonquad_obj_exprs = []
        self._approx_obj_exprs = []

        self._nonlin_cnt_exprs = []

        # These are lists of OSQPQuadraticObj's, OSQPLinearObj's and OSQPLinearConstraints
        # that will directly be used to construct the P, q and A matrices that define
        # the final QP to be solved.
        self._osqp_quad_objs = []
        self._osqp_lin_objs = []
        self._osqp_lin_cnt_exprs = []

        # list of constraints that will hold the hinge constraints
        # for each non-linear constraint, is a pair of constraints
        # for an eq constraint
        self.hinge_created = False

        self._penalty_exprs = []
        self._osqp_penalty_cnts = []
        self._osqp_penalty_exprs = []

        # group-id (str) -> cnt-set (set of constraints)
        self._cnt_groups = defaultdict(set)
        self._cnt_groups_overlap = defaultdict(set)
        self._penalty_groups = []
        self.nonconverged_groups = []
        self.gid2ind = {}

    def add_obj_expr(self, bound_expr):
        """
        Adds a bound expression (bound_expr) to the objective. If the objective
        is quadratic, is it added to _quad_obj_exprs. Otherwise, it is added
        to self._nonquad_obj_exprs.

        bound_expr's var is added to self._vars so that a trust region can be
        added to var.
        """
        expr = bound_expr.expr
        if isinstance(expr, sco_osqp_expr.AffExpr) or isinstance(
            expr, sco_osqp_expr.QuadExpr
        ):
            self._quad_obj_exprs.append(bound_expr)
        else:
            self._nonquad_obj_exprs.append(bound_expr)
        self.add_var(bound_expr.var)

    def add_var(self, var):
        self._vars.add(var)

    def add_osqp_var(self, osqp_var):
        self._osqp_vars.add(osqp_var)

    def add_cnt_expr(self, bound_expr, group_ids=None):
        """
        Adds a bound expression (bound_expr) to the problem's constraints.
        If the constraint is linear, it is added directly to the model.
        Otherwise, the constraint is added by appending bound_expr to
        self._nonlin_cnt_exprs.

        bound_expr's var is added to self._vars so that a trust region can be
        added to var.
        """
        comp_expr = bound_expr.expr
        expr = comp_expr.expr
        var = bound_expr.var
        assert isinstance(comp_expr, sco_osqp_expr.CompExpr)
        if isinstance(expr, sco_osqp_expr.AffExpr):
            if isinstance(comp_expr, sco_osqp_expr.EqExpr):
                self._add_osqp_cnt_from_aff_expr(expr, var, "eq", comp_expr.val)
            elif isinstance(comp_expr, sco_osqp_expr.LEqExpr):
                self._add_osqp_cnt_from_aff_expr(expr, var, "leq", comp_expr.val)
        else:
            self._nonlin_cnt_exprs.append(bound_expr)
            self._reset_hinge_cnts()

            if group_ids is None:
                group_ids = ["all"]
            for gid in group_ids:
                self._cnt_groups[gid].add(bound_expr)
                for other in group_ids:
                    if other == gid:
                        continue
                    self._cnt_groups_overlap[gid].add(other)

        self.add_var(var)

    def optimize(self, 
                 add_convexified_terms=False, 
                 osqp_eps_abs=1e-06, 
                 osqp_eps_rel=1e-09, 
                 osqp_max_iter=osqp_utils.DEFAULT_MAX_ITER, 
                 verbose=False):
        """
        Calls the OSQP optimizer on the current QP approximation with a given
        penalty coefficient. Note that add_convexified_terms is a convenience
        boolean useful to toggle whether or not self._osqp_penalty_exprs and
        self._osqp_penalty_cnts are included in the optimization problem
        """
        if not add_convexified_terms:
            solve_res, var_to_index_dict = osqp_utils.optimize(
                self._osqp_vars,
                self._vars,
                self._osqp_quad_objs,
                self._osqp_lin_objs,
                self._osqp_lin_cnt_exprs,
                osqp_eps_abs,
                osqp_eps_rel,
                osqp_max_iter,
                verbose=verbose,
            )
        else:
            cnt_exprs = self._osqp_lin_cnt_exprs[:]
            for penalty_cnt_list in self._osqp_penalty_cnts:
                cnt_exprs.extend(penalty_cnt_list)

            solve_res, var_to_index_dict = osqp_utils.optimize(
                self._osqp_vars,
                self._vars,
                self._osqp_quad_objs,
                self._osqp_lin_objs + self._osqp_penalty_exprs,
                cnt_exprs,
                osqp_eps_abs,
                osqp_eps_rel,
                osqp_max_iter,
                verbose=verbose,
            )

        # If the solve failed, just return False
        if solve_res.info.status_val not in [1, 2]:
            return False

        # If the solve succeeded, update all the variables with these new values, then
        # run he callback before returning true
        osqp_utils.update_osqp_vars(var_to_index_dict, solve_res.x)
        self._update_vars()
        self._callback() # TODO: Modify to get the visualizer in a better spot.
        return True

    def _reset_hinge_cnts(self):
        # reset the hinge_cnts
        self.hinge_created = False

    def _add_osqp_objs_and_cnts_from_expr(self, bound_expr):
        """
        Uses AffExpr, QuadExpr, HingeExpr and AbsExpr to extract
        OSQP solver compatible data structures. Depending on the expression type,
        appends elements to self._osqp_quad_objs, or self._osqp_lin_objs.
        """
        expr = bound_expr.expr
        var = bound_expr.var

        if isinstance(expr, sco_osqp_expr.AffExpr):
            self._add_to_lin_objs_and_cnts_from_aff_expr(expr, var)
        elif isinstance(expr, sco_osqp_expr.QuadExpr):
            self._add_to_quad_and_lin_objs_from_quad_expr(expr, var)
        elif isinstance(expr, sco_osqp_expr.HingeExpr):
            self._add_to_lin_objs_and_cnts_from_hinge_expr(expr, var)
        elif isinstance(expr, sco_osqp_expr.AbsExpr):
            self._add_to_lin_objs_and_cnts_from_abs_expr(expr, var)
        elif isinstance(expr, sco_osqp_expr.CompExpr):
            raise Exception(
                "Comparison Expressions cannot be converted to \
                OSQP problem objectives; use _add_osqp_cnt_from_aff_expr \
                instead"
            )
        else:
            raise Exception(
                "This type of Expression cannot be converted to\
                an OSQP objective."
            )

    def _add_to_lin_objs_and_cnts_from_aff_expr(self, aff_expr, var):
        osqp_vars = var.get_osqp_vars()
        A_mat = aff_expr.A

        for i in range(A_mat.shape[0]):
            (inds,) = np.nonzero(A_mat[i, :])
            for coeff, osqp_var in zip(
                A_mat[i, inds].tolist(), osqp_vars[inds, 0].tolist()
            ):
                self._osqp_penalty_exprs.append(OSQPLinearObj(osqp_var, coeff))

    def _add_to_lin_objs_and_cnts_from_hinge_expr(self, hinge_expr, var):
        aff_expr = hinge_expr.expr
        assert isinstance(aff_expr, sco_osqp_expr.AffExpr)
        osqp_vars = var.get_osqp_vars()
        A_mat = aff_expr.A
        b_vec = aff_expr.b

        hinge = self.create_pos_osqp_var_arr((A_mat.shape[0], 1))
        for _, hinge_var in np.ndenumerate(hinge):
            self._osqp_penalty_exprs.append(OSQPLinearObj(hinge_var, 1.0))

        cnts_list = []
        for i in range(A_mat.shape[0]):
            (inds,) = np.nonzero(A_mat[i, :])
            # since this is a hinge constraint, the lower bound is -inf
            curr_lb = -np.inf
            # the constraint expr is:
            # curr_lb <= osqp_vars[inds, 0] * A_mat[i, inds] <= hinge[i] - b_vec[i]
            # which can be reformulated as:
            # curr_lb <= osqp_vars[inds, 0] * A_mat[i, inds] - hinge[i] <= -b_vec[i]
            # since curr_lb is -inf
            curr_ub = -b_vec[i]
            cnt_vars = np.concatenate((osqp_vars[inds, 0], hinge[i]))
            cnt_coeffs = np.concatenate((A_mat[i, inds], np.array([-1.0])))
            curr_cnt_expr = OSQPLinearConstraint(cnt_vars, cnt_coeffs, curr_lb, curr_ub)
            cnts_list.append(curr_cnt_expr)

        self._osqp_penalty_cnts.append(cnts_list)

    def _add_to_lin_objs_and_cnts_from_abs_expr(self, abs_expr, var):
        aff_expr = abs_expr.expr
        assert isinstance(aff_expr, sco_osqp_expr.AffExpr)
        A_mat = aff_expr.A
        b_vec = aff_expr.b
        pos = self.create_pos_osqp_var_arr((A_mat.shape[0], 1))
        neg = self.create_pos_osqp_var_arr((A_mat.shape[0], 1))

        # First, add all the objective terms
        for pos_var in pos.flat:
            self._osqp_penalty_exprs.append(OSQPLinearObj(pos_var, 1.0))
        for neg_var in neg.flat:
            self._osqp_penalty_exprs.append(OSQPLinearObj(neg_var, 1.0))

        # Next, add the constraints
        osqp_vars = var.get_osqp_vars()
        A_mat = aff_expr.A
        b_vec = aff_expr.b

        cnts_list = []
        for i in range(A_mat.shape[0]):
            (inds,) = np.nonzero(A_mat[i, :])
            # the constraint expr is:
            # pos_var[i] - neg_var[i] <= osqp_vars[inds, 0] * A_mat[i, inds] + b_vec[i] <= pos_var[i] - neg_var[i]
            # which can be reformulated as:
            # -b_vec[i] <= osqp_vars[inds, 0] * A_mat[i, inds] - pos_var[i] + neg_var[i] <= -b_vec[i]
            curr_ub = -b_vec[i]
            curr_lb = -b_vec[i]
            cnt_vars = np.concatenate((osqp_vars[inds, 0], pos[i], neg[i]))
            cnt_coeffs = np.concatenate(
                (A_mat[i, inds], np.array([-1.0]), np.array([1.0]))
            )
            curr_cnt_expr = OSQPLinearConstraint(cnt_vars, cnt_coeffs, curr_lb, curr_ub)
            cnts_list.append(curr_cnt_expr)

        self._osqp_penalty_cnts.append(cnts_list)

    def _add_osqp_cnt_from_aff_expr(self, aff_expr, var, cnt_type, cnt_val):
        """
        Uses aff_expr to create OSQPLinearConstraints of cnt_type that are then
        appended to self._osqp_lin_cnt_exprs
        """
        osqp_vars = var.get_osqp_vars()
        A_mat = aff_expr.A
        b_vec = aff_expr.b
        for i in range(A_mat.shape[0]):
            (inds,) = np.nonzero(A_mat[i, :])
            # If the constraint to be added is an equality constraint,
            # compute the upper and lower bounds
            if cnt_type == "eq":
                # the upper and lower bounds must be equal, and they must be
                # whatever the cnt_val was minus the constant term
                curr_lb = cnt_val[i] - b_vec[i]
                curr_ub = cnt_val[i] - b_vec[i]
            elif cnt_type == "leq":
                # only the upper bound needs to be set; the lower bound is negative
                # infinity
                curr_lb = -np.inf
                curr_ub = cnt_val[i] - b_vec[i]
            else:
                raise NotImplementedError

            curr_cnt_expr = OSQPLinearConstraint(
                osqp_vars[inds, 0], A_mat[i, inds], curr_lb, curr_ub
            )

            self._osqp_lin_cnt_exprs.append(curr_cnt_expr)

    def _add_to_quad_and_lin_objs_from_quad_expr(self, quad_expr, var):
        x = var.get_osqp_vars()
        Q = quad_expr.Q
        rows, cols = x.shape
        assert cols == 1
        inds = np.nonzero(Q)
        coeffs = Q[inds]  # No need to multiply by 2 because OSQP expects 0.5*x.T*Q*x
        v1 = x[inds[0], 0]
        v2 = x[inds[1], 0]
        # Create the new QuadraticObj term and append it to the problem's running list of
        # such terms
        self._osqp_quad_objs.append(OSQPQuadraticObj(v1, v2, coeffs))
        inds = np.nonzero(quad_expr.A)
        lin_coeffs = quad_expr.A[inds]
        # Because quad_expr.A is of shape (1,2), inds[1] corresponds to the nonzero
        # vars
        lin_vars = x[inds[1], 0]
        assert lin_coeffs.shape == lin_vars.shape
        for lin_var, lin_coeff in zip(lin_vars.tolist(), lin_coeffs.tolist()):
            self._osqp_lin_objs.append(OSQPLinearObj(lin_var, lin_coeff))

    def find_closest_feasible_point(self):
        """
        Finds the closest point (l2 norm) to the initialization that satisfies
        the linear constraints.
        """
        for var in self._vars:
            osqp_vars = var.get_osqp_vars()
            val = var.get_value()
            if val is not None:
                assert osqp_vars.shape == val.shape
                inds = np.where(~np.isnan(val))
                val = val[inds]
                nonnan_osqp_vars = osqp_vars[inds]
                val_arr = val.flatten()
                for i, nonnan_osqp_var in enumerate(
                    nonnan_osqp_vars.flatten().tolist()
                ):
                    # Create the correct quad and lin objectives!
                    self._osqp_lin_objs.append(
                        OSQPLinearObj(nonnan_osqp_var, -2.0 * val_arr[i])
                    )

                    self._osqp_quad_objs.append(
                        OSQPQuadraticObj(
                            np.array([nonnan_osqp_var]),
                            np.array([nonnan_osqp_var]),
                            np.array([2.0]),
                        )
                    )

        return self.optimize()

    def update_obj(self, penalty_coeff=0.0):
        self._reset_osqp_objs()
        self._lazy_spawn_osqp_cnts()

        for bound_expr in self._quad_obj_exprs + self._approx_obj_exprs:
            self._add_osqp_objs_and_cnts_from_expr(bound_expr)

        for i, bound_expr in enumerate(self._penalty_exprs):
            self._update_nonlin_cnt_and_add_to_qp(bound_expr, i)

        for lin_obj in self._osqp_penalty_exprs:
            lin_obj.coeff = lin_obj.coeff * penalty_coeff
            self._osqp_lin_objs.append(lin_obj)

    def _reset_osqp_objs(self):
        """Resets the quadratic and linear objectives in preparation for the
        definition of a new OSQP problem"""
        self._osqp_quad_objs = []
        self._osqp_lin_objs = []

    def _lazy_spawn_osqp_cnts(self):
        if not self.hinge_created:
            self._osqp_penalty_cnts = []
            self._osqp_penalty_exprs = []
            self._osqp_nz = []
            for bound_expr in self._penalty_exprs:
                self._osqp_nz.append(np.nonzero(bound_expr.expr.expr.A))
                # The below line populates self._osqp_penalty_exprs and
                # self._osqp_penalty_cnts
                self._add_osqp_objs_and_cnts_from_expr(bound_expr)
            self.hinge_created = True

    def create_pos_osqp_var_arr(self, shape):
        """
        Returns a numpy array of new, unused positive OSQP variables.
        """
        osqp_var_arr = np.empty(shape, dtype=object)
        for x in np.nditer(osqp_var_arr, op_flags=["readwrite"], flags=["refs_ok"]):
            # Create a new variable that's bound between 0 and np.inf, and name it so that
            # it so it starts with a z and will get sorted to the end
            new_pos_var = OSQPVar("z+_pos_osqp_var", 0.0, np.inf, 0.0)
            # Add it to the set keeping track of all OSQP vars
            self._osqp_vars.add(new_pos_var)
            x[...] = new_pos_var
        return osqp_var_arr

    # @profile
    def _update_nonlin_cnt_and_add_to_qp(self, bexpr, ind):
        """Updates all non-linear constraints and adds to new (approximated) linear
        constraints to _osqp_lin_cnt_exprs"""

        expr, var = bexpr.expr, bexpr.var
        if isinstance(expr, sco_osqp_expr.HingeExpr) or isinstance(
            expr, sco_osqp_expr.AbsExpr
        ):
            aff_expr = expr.expr
            assert isinstance(aff_expr, sco_osqp_expr.AffExpr)
            A, b = aff_expr.A, aff_expr.b
            cnts = self._osqp_penalty_cnts[ind]
            old_nz = self._osqp_nz[ind]
            osqp_vars = var.get_osqp_vars()
            nz = np.nonzero(A)

            for i in range(A.shape[0]):
                # if lb == ub, then it is an equality constraint, so
                # we need to set both lb and ub to -b[i, 0]
                if cnts[i].lb == cnts[i].ub:
                    cnts[i].lb = -b[i, 0]
                    cnts[i].ub = -b[i, 0]
                # otherwise, it is a LEQ constraint and so
                # we only need to update the ub
                else:
                    cnts[i].ub = -b[i, 0]

            for idx in range(old_nz[0].shape[0]):
                i, j = old_nz[0][idx], old_nz[1][idx]
                # if osqp_vars[j, 0] is in cnts[i].osqp_vars, then modify
                # cnts[i].osqp_vars to set the associated coefficient
                # to 0
                if osqp_vars[j, 0] in cnts[i].osqp_vars:
                    rep_i = np.where(cnts[i].osqp_vars == osqp_vars[j, 0])[0][0]
                    cnts[i].coeffs[rep_i] = 0.0

            for idx in range(nz[0].shape[0]):
                i, j = nz[0][idx], nz[1][idx]
                # if osqp_vars[j, 0] is in cnts[i].osqp_vars, then modify
                # cnts[i].osqp_vars to set the associated coefficient
                # to A[i, j]
                if osqp_vars[j, 0] in cnts[i].osqp_vars:
                    rep_i = np.where(cnts[i].osqp_vars == osqp_vars[j, 0])[0][0]
                    cnts[i].coeffs[rep_i] = A[i, j]

            self._osqp_nz[ind] = nz

            for cnt in cnts:
                self._osqp_lin_cnt_exprs.append(cnt)

        else:
            raise NotImplementedError

    def add_trust_region(self, trust_region_size):
        """
        Adds the trust region for every variable
        """
        for var in self._vars:
            var.add_trust_region(trust_region_size)

    # @profile
    def convexify(self):
        """
        Convexifies the optimization problem by computing a QP approximation
        A quadratic approximation of the non-quadratic objective terms
        (self._nonquad_obj_exprs) is saved in self._approx_obj_exprs.
        The penalty approximation of the non-linear constraints
        (self._nonlin_cnt_exprs) is saved in self._penalty_exprs
        """
        # if len(self._nonquad_obj_exprs) > 0:
        #     import ipdb; ipdb.set_trace()
        self._approx_obj_exprs = [
            bexpr.convexify(degree=2) for bexpr in self._nonquad_obj_exprs
        ]
        self._penalty_exprs = [
            bexpr.convexify(degree=1) for bexpr in self._nonlin_cnt_exprs
        ]
        self._penalty_groups = []
        gids = sorted(self._cnt_groups.keys())
        self.gid2ind = {}
        for i, gid in enumerate(gids):
            self.gid2ind[gid] = i
            cur_bexprs = [bexpr.convexify(degree=1) for bexpr in self._cnt_groups[gid]]
            self._penalty_groups.append(cur_bexprs)

    # #@profile
    def get_value(self, penalty_coeff, vectorize=False):
        """
        Returns the current value of the penalty objective.
        The penalty objective is computed by summing up all the values of the
        quadratic objective expressions (self._quad_obj_exprs), the
        non-quadratic objective expressions and the penalty coeff multiplied
        by the constraint violations (computed using _nonlin_cnt_exprs)

        if vectorize=True, then this returns a vector of constraint
        violations -- 1 per group id.
        """
        if vectorize:
            gids = sorted(self._cnt_groups.keys())
            value = np.zeros(len(gids))
            for i, gid in enumerate(gids):
                value[i] = np.sum(
                    np.sum(
                        [
                            np.sum(self._compute_cnt_violation(bexpr))
                            for bexpr in self._cnt_groups[gid]
                        ]
                    )
                )
            return value
        else:
            value = 0.0
            for bound_expr in self._quad_obj_exprs + self._nonquad_obj_exprs:
                value += np.sum(np.sum(bound_expr.eval()))
            for bound_expr in self._nonlin_cnt_exprs:
                cnt_vio = self._compute_cnt_violation(bound_expr)
                value += penalty_coeff * np.sum(cnt_vio)

        return value

    # @profile
    def _compute_cnt_violation(self, bexpr):
        comp_expr = bexpr.expr
        var_val = bexpr.var.get_value()
        if isinstance(comp_expr, sco_osqp_expr.EqExpr):
            return np.absolute(comp_expr.expr.eval(var_val) - comp_expr.val)
        elif isinstance(comp_expr, sco_osqp_expr.LEqExpr):
            v = comp_expr.expr.eval(var_val) - comp_expr.val
            zeros = np.zeros(v.shape)
            return np.maximum(v, zeros)

    def get_max_cnt_violation(self):
        """
        Returns the the maximum amount a non-linear constraint is violated.
        Linear constraints are assumed to be satisfied because they are added
        directly to the model and QP solvers can deal with them.
        """
        max_vio = 0.0
        for bound_expr in self._nonlin_cnt_exprs:
            cnt_vio = self._compute_cnt_violation(bound_expr)
            cnt_max_vio = np.amax(cnt_vio)
            max_vio = np.maximum(max_vio, cnt_max_vio)
        return max_vio

    def get_approx_value(self, penalty_coeff, vectorize=False):
        """
        Returns the current value of the penalty QP approximation by summing
        up the expression values for the quadratic objective terms
        (_quad_obj_exprs), the quadratic approximation of the non-quadratic
        terms (_approx_obj_exprs) and the penalty terms (_penalty_exprs).
        Note that this approximate value is computed with respect to when the
        last convexification was performed.

        if vectorize=True, then this returns a vector of constraint
        violations -- 1 per group id.
        """
        if vectorize:
            value = np.zeros(len(self._penalty_groups))
            for i, bexprs in enumerate(self._penalty_groups):
                x = np.array([np.sum(bexpr.eval()) for bexpr in bexprs])
                value[i] = np.sum(x.flatten())
            return value

        value = 0.0
        for bound_expr in self._quad_obj_exprs + self._approx_obj_exprs:
            value += np.sum(np.sum(bound_expr.eval()))
        for bound_expr in self._penalty_exprs:
            value += penalty_coeff * np.sum(bound_expr.eval())

        return value

    def _update_vars(self):
        for var in self._vars:
            var.update()

    def batch_add_lin_cnts(self, list_of_lin_cnts):
        self._osqp_lin_cnt_exprs.extend(list_of_lin_cnts)

    def save(self):
        """
        Saves the problem's current state by saving the values of all the
        variables.
        """
        for var in self._vars:
            var.save()

    def restore(self):
        """
        Restores the problem's state to the problem's saved state
        """
        for var in self._vars:
            var.restore()



import numpy as np


class Variable(object):
    """
    Variable

    Manages low-level variables by maintaining an ordering of these variables,
    """

    def __init__(self, osqp_vars, value=None):
        """
        _osqp_vars: Numpy array of OSQPVars that represent the variables
        _value: Numpy array of the current values of these variables
        _saved_value: saved value of this variable
        """
        assert isinstance(osqp_vars, np.ndarray)
        assert len(osqp_vars) > 0
        self._osqp_vars = osqp_vars.copy()
        if value is not None:
            assert osqp_vars.shape == value.shape
            assert isinstance(value, np.ndarray)
            self._value = value.copy()
        else:
            self._value = None
        self._saved_value = None

    def get_osqp_vars(self):
        return self._osqp_vars

    def get_value(self):
        if self._value is not None:
            return self._value.copy()
        else:
            return None

    def add_trust_region(self, trust_box_size):
        """
        Adds a trust region around the saved value (self._saved_value) by
        changing the upper and lower bounds of the underlying OSQP variables
        """
        assert self._saved_value is not None
        for index, osqp_var in np.ndenumerate(self._osqp_vars):
            osqp_var.set_lower_bound(self._saved_value[index] - trust_box_size)
            osqp_var.set_upper_bound(self._saved_value[index] + trust_box_size)

    def update(self):
        """
        If the OSQP variables have valid values, update self._value to reflect
        the values in these variables.
        """
        value = np.zeros(self._osqp_vars.shape)
        for index, osqp_var in np.ndenumerate(self._osqp_vars):
            if osqp_var.val is not None:
                value[index] = osqp_var.val
            else:
                raise ValueError(
                    f"The variable {osqp_var.var_name} does not have a legitimate value"
                )
        self._value = value

    def save(self):
        """
        Save the current value.
        """
        assert not np.any(np.isnan(self._value.copy()))
        self._saved_value = self._value.copy()

    def restore(self):
        """
        Restore value to the saved value.
        """
        self._value = self._saved_value.copy()



import time

import numpy as np

import sco_py.sco_osqp.osqp_utils as osqp_utils


class Solver(object):
    """
    SCO Solver
    """

    def __init__(self):
        """
        values taken from Pieter Abbeel's CS287 hw3 q2 penalty_sqp.m file
        """
        self.improve_ratio_threshold = 0.25
        self.min_trust_region_size = 1e-4
        # self.min_approx_improve = 1e-4
        self.min_approx_improve = 1e-8
        self.max_iter = 50
        self.trust_shrink_ratio = 0.1
        self.trust_expand_ratio = 1.5
        self.cnt_tolerance = 1e-4
        self.max_merit_coeff_increases = 1
        self.merit_coeff_increase_ratio = 1e1
        self.initial_trust_region_size = 1
        self.initial_penalty_coeff = 1e3

    def solve(self, 
              prob, 
              method=None, 
              tol=None, 
              verbose=False,
              osqp_eps_abs=1e-06, 
              osqp_eps_rel=1e-09, 
              osqp_max_iter=osqp_utils.DEFAULT_MAX_ITER):
        """
        Returns whether solve succeeded.

        Given a sco (sequential convex optimization) problem instance, solve
        using specified method to find a solution. If the specified method
        doesn't exist, an exception is thrown.
        """
        if tol is not None:
            self.min_trust_region_size = tol
            self.min_approx_improve = tol
            self.cnt_tolerance = tol

        if method == "penalty_sqp":
            return self._penalty_sqp(prob, verbose=verbose,\
                osqp_eps_abs=osqp_eps_abs, osqp_eps_rel=osqp_eps_rel, osqp_max_iter=osqp_max_iter)
        else:
            raise Exception("This method is not supported.")

    # @profile
    def _penalty_sqp(self, 
                     prob, 
                     verbose=False,
                     osqp_eps_abs=2e-06, 
                     osqp_eps_rel=1e-09, 
                     osqp_max_iter=osqp_utils.DEFAULT_MAX_ITER):
        """
        Return true is the penalty sqp method succeeds.
        Uses Penalty Sequential Quadratic Programming to solve the problem
        instance.
        """
        start = time.time()
        trust_region_size = self.initial_trust_region_size
        penalty_coeff = self.initial_penalty_coeff

        if not prob.find_closest_feasible_point():
            return False

        for i in range(self.max_merit_coeff_increases):
            success = self._min_merit_fn(
                prob, penalty_coeff, trust_region_size, verbose=verbose,\
                    osqp_eps_abs=osqp_eps_abs, osqp_eps_rel=osqp_eps_rel,\
                        osqp_max_iter=osqp_max_iter
            )
            if verbose:
                print("\n")

            if prob.get_max_cnt_violation() > self.cnt_tolerance:
                penalty_coeff = penalty_coeff * self.merit_coeff_increase_ratio
                trust_region_size = self.initial_trust_region_size
            else:
                end = time.time()
                if verbose:
                    print("sqp time: ", end - start)
                return success
        end = time.time()
        if verbose:
            print("sqp time: ", end - start)
        return False

    # @profile
    def _min_merit_fn(self, prob, penalty_coeff, trust_region_size, verbose=False,\
        osqp_eps_abs=1e-06, osqp_eps_rel=1e-09, osqp_max_iter=osqp_utils.DEFAULT_MAX_ITER):
        """
        Minimize merit function for penalty sqp.
        Returns true if the merit function is minimized successfully.
        """
        sqp_iter = 1

        while True:
            if verbose:
                print(("  sqp_iter: {0}".format(sqp_iter)))

            prob.convexify()
            prob.update_obj(penalty_coeff)
            merit = prob.get_value(penalty_coeff)
            merit_vec = prob.get_value(penalty_coeff, True)
            prob.save()

            while True:
                if verbose:
                    print(("    trust region size: {0}".format(trust_region_size)))
                prob.add_trust_region(trust_region_size)
                _ = prob.optimize(osqp_eps_abs=osqp_eps_abs, 
                                  osqp_eps_rel=osqp_eps_rel,\
                                  osqp_max_iter=osqp_max_iter,
                                  verbose=verbose)
                model_merit = prob.get_approx_value(penalty_coeff)
                model_merit_vec = prob.get_approx_value(penalty_coeff, True)
                new_merit = prob.get_value(penalty_coeff)

                approx_merit_improve = merit - model_merit
                if not approx_merit_improve:
                   approx_merit_improve += 1e-12

                # we converge if one of the violated constraint groups
                # is below the minimum improvement
                approx_improve_vec = merit_vec - model_merit_vec
                violated = merit_vec > self.cnt_tolerance
                if approx_improve_vec.shape == (0,):
                    approx_improve_vec = np.array([approx_merit_improve])
                    violated = approx_improve_vec > -np.inf

                exact_merit_improve = merit - new_merit

                merit_improve_ratio = exact_merit_improve / approx_merit_improve

                if verbose:
                    print(
                        (
                            "      merit: {0}. model_merit: {1}. new_merit: {2}".format(
                                merit, model_merit, new_merit
                            )
                        )
                    )
                    print(
                        (
                            "      approx_merit_improve: {0}. exact_merit_improve: {1}. merit_improve_ratio: {2}".format(
                                approx_merit_improve,
                                exact_merit_improve,
                                merit_improve_ratio,
                            )
                        )
                    )

                if self._bad_model(approx_merit_improve):
                    if verbose:
                        print(
                            (
                                "Approximate merit function got worse ({0})".format(
                                    approx_merit_improve
                                )
                            )
                        )
                        print(
                            "Either convexification is wrong to zeroth order, or you're in numerical trouble."
                        )
                    prob.restore()
                    return False

                if self._y_converged(approx_merit_improve):
                    if verbose:
                        print("Converged: y tolerance")
                    prob.restore()
                    return True

                # we converge if one of the violated constraint groups
                # is below the minimum improvement and none of its overlapping
                # groups are making progress
                prob.nonconverged_groups = []
                for gid, idx in prob.gid2ind.items():
                    if (
                        violated[idx]
                        and approx_improve_vec[idx] < self.min_approx_improve
                    ):
                        overlap_improve = False
                        for gid2 in prob._cnt_groups_overlap[gid]:
                            if (
                                approx_improve_vec[prob.gid2ind[gid2]]
                                > self.min_approx_improve
                            ):
                                overlap_improve = True
                                break
                        if overlap_improve:
                            continue
                        prob.nonconverged_groups.append(gid)
                if len(prob.nonconverged_groups) > 0:
                    if verbose:
                        print("Converged: y tolerance")
                    prob.restore()
                    # store the failed groups into the prob

                    for i, g in enumerate(sorted(prob._cnt_groups.keys())):
                        if violated[i] and self._y_converged(approx_improve_vec[i]):
                            prob.nonconverged_groups.append(g)
                    return True

                if self._shrink_trust_region(exact_merit_improve, merit_improve_ratio):
                    prob.restore()
                    if verbose:
                        print("Shrinking trust region")
                    trust_region_size = trust_region_size * self.trust_shrink_ratio
                else:
                    if verbose:
                        print("Growing trust region")
                    trust_region_size = trust_region_size * self.trust_expand_ratio
                    break  # from trust region loop

                if self._x_converged(trust_region_size):
                    if verbose:
                        print("Converged: x tolerance")
                    return True

            sqp_iter = sqp_iter + 1

    def _bad_model(self, approx_merit_improve):
        """
        Returns true if the approx_merit_improve is too low which suggests that
        either the convexification is wrong to the zeroth order or there are
        numerical problems.
        """
        return approx_merit_improve < -1e-5

    def _shrink_trust_region(self, exact_merit_improve, merit_improve_ratio):
        """
        Returns true if the trust region should shrink (exact merit improve is negative or the merit improve ratio is too low)
        """
        return (exact_merit_improve < 0) or (
            merit_improve_ratio < self.improve_ratio_threshold
        )

    def _x_converged(self, trust_region_size):
        """
        Returns true if the variable values has converged (trust_region size is
        smaller than the minimum trust region size)
        """
        return trust_region_size < self.min_trust_region_size

    def _y_converged(self, approx_merit_improve):
        """
        Returns true if the approx_merit has converged (approx_merit_improve <
        min_approx_merit_improve)
        """
        return approx_merit_improve < self.min_approx_improve



from typing import List

import numpy as np
import osqp
import scipy

from sco_py.sco_osqp.variable import Variable


DEFAULT_MAX_ITER = int(1e05)

class OSQPVar(object):
    """
    A class representing a variable within OSQP that will be used to construct the
    QP that will be passed to the solver. The lb and ub are used to enforce a
    trust region on each variable
    """

    def __init__(self, var_name, lb=-np.inf, ub=np.inf, val=None):
        self.var_name = var_name
        self._lower_bound = lb
        self._upper_bound = ub
        self.val = val

    def __lt__(self, other_osqp_var):
        return self.var_name < other_osqp_var.var_name

    def __repr__(self):
        return f"OSQPVar with name {self.var_name}"
        # return f"OSQPVar with name {self.var_name}, lb={self._lower_bound}, ub={self._upper_bound}, val={self.val}"

    def get_lower_bound(self):
        return self._lower_bound

    def set_lower_bound(self, lb_val):
        assert isinstance(lb_val, float)
        assert not np.isnan(lb_val)
        self._lower_bound = lb_val

    def get_upper_bound(self):
        return self._upper_bound

    def set_upper_bound(self, ub_val):
        assert isinstance(ub_val, float)
        assert not np.isnan(ub_val)
        self._upper_bound = ub_val


class OSQPLinearObj(object):
    def __init__(self, osqp_var, coeff):
        """
        A wrapper class for a linear objective. The particular constraint represented is:

        coeff * osqp_var
        """
        self.osqp_var = osqp_var
        self.coeff = coeff

    def __repr__(self):
        return f"OSQPLinearObj with osqp_var={self.osqp_var}, coeff={self.coeff}"

    def get_all_vars(self):
        return [self.osqp_var]


class OSQPQuadraticObj(object):
    def __init__(self, osqp_vars1, osqp_vars2, coeffs):
        """
        A wrapper class for a quadratic objective. The particular objective represented is:

        0.5 * np.linalg.multi_dot(self.osqp_vars1, self.osqp_vars2, self.coeffs)
        """
        assert osqp_vars1.shape == osqp_vars2.shape == coeffs.shape
        # Together with the above condition, this implicitly implies all these arrays must
        # have shape of length 1
        assert len(osqp_vars1.shape) == 1
        self.osqp_vars1 = osqp_vars1
        self.osqp_vars2 = osqp_vars2
        self.coeffs = coeffs

    def __repr__(self):
        return f"Quadratic Objective with osqp_vars1={self.osqp_vars1}, osqp_vars2={self.osqp_vars2}, coeffs={self.coeffs}"

    def get_all_vars(self):
        return self.osqp_vars1.tolist() + self.osqp_vars2.tolist()


class OSQPLinearConstraint(object):
    def __init__(self, osqp_vars, coeffs, lb, ub):
        """
        A wrapper class for a linear constraint. The particular constraint represented is:

        lb <= np.dot(self.osqp_vars, self.coeffs) <= ub
        """
        assert osqp_vars.shape == coeffs.shape
        self.osqp_vars = osqp_vars
        self.coeffs = coeffs
        self.lb = lb
        self.ub = ub

    def __repr__(self):
        return f"OSQPLinearConstraint with osqp_vars={self.osqp_vars}, coeffs={self.coeffs}, lb = {self.lb}, ub = {self.ub}"

    def get_all_vars(self):
        return self.osqp_vars.tolist()

# @profile
def optimize(
    osqp_vars: List[OSQPVar],
    _sco_vars: List[Variable],
    osqp_quad_objs: List[OSQPQuadraticObj],
    osqp_lin_objs: List[OSQPLinearObj],
    osqp_lin_cnt_exprs: List[OSQPLinearConstraint],
    eps_abs: float = 1e-06,
    eps_rel: float = 1e-09,
    max_iter: int = DEFAULT_MAX_ITER,
    verbose: bool = False,
):
    """
    Calls the OSQP optimizer on the current QP approximation with a given
    penalty coefficient.
    """

    # First, we need to setup the problem as described here: https://osqp.org/docs/solver/index.html
    # Specifically, we need to start by constructing the x vector that contains all the
    # OSQPVars that are part of the QP. This will take the form of a mapping from OSQPVar to
    # index within the x vector.
    var_to_index_dict = {}
    osqp_var_list = list(osqp_vars)
    # Make sure to sort this list to get a canonical ordering of variables to make 
    # matrix construction easier
    osqp_var_list.sort()  
    for idx, osqp_var in enumerate(osqp_var_list):
        var_to_index_dict[osqp_var] = idx
    num_osqp_vars = len(osqp_vars)

    # Construct the q-vector by looping through all the linear objectives
    q_vec = np.zeros(num_osqp_vars)
    for lin_obj in osqp_lin_objs:
        q_vec[var_to_index_dict[lin_obj.osqp_var]] += lin_obj.coeff

    # Next, construct the P-matrix by looping through all quadratic objectives

    # Since P must be upper-triangular, the shape must be (num_osqp_vars, num_osqp_vars)
    P_mat = np.zeros((num_osqp_vars, num_osqp_vars))
    for quad_obj in osqp_quad_objs:
        for i in range(quad_obj.coeffs.shape[0]):
            idx2 = var_to_index_dict[quad_obj.osqp_vars1[i]]
            idx1 = var_to_index_dict[quad_obj.osqp_vars2[i]]
            if idx1 > idx2:
                P_mat[idx2, idx1] += 0.5 * quad_obj.coeffs[i]
            elif idx1 < idx2:
                P_mat[idx1, idx2] += 0.5 * quad_obj.coeffs[i]
            else:
                P_mat[idx1, idx2] += quad_obj.coeffs[i]

    # Next, setup the A-matrix and l and u vectors
    A_mat = np.zeros((num_osqp_vars + len(osqp_lin_cnt_exprs), num_osqp_vars))
    l_vec = np.zeros(num_osqp_vars + len(osqp_lin_cnt_exprs))
    u_vec = np.zeros(num_osqp_vars + len(osqp_lin_cnt_exprs))

    # First add all the linear constraints
    # However, note that this isn't entirely straightforward: some 
    row_num = 0
    for lin_constraint in osqp_lin_cnt_exprs:
        l_vec[row_num] = lin_constraint.lb
        u_vec[row_num] = lin_constraint.ub
        for i in range(lin_constraint.coeffs.shape[0]):
            # if var_to_index_dict[lin_constraint.osqp_vars[i]] == 193:
            #     import ipdb; ipdb.set_trace()
            A_mat[
                row_num, var_to_index_dict[lin_constraint.osqp_vars[i]]
            ] = lin_constraint.coeffs[i]
        row_num += 1

    # Then, add the trust regions for every variable as constraints
    for osqp_var in osqp_vars:
        A_mat[row_num, var_to_index_dict[osqp_var]] = 1.0
        l_vec[row_num] = osqp_var.get_lower_bound()
        u_vec[row_num] = osqp_var.get_upper_bound()
        row_num += 1

    # Finally, construct the matrices and call the OSQP Solver!
    P_mat_sparse = scipy.sparse.csc_matrix(P_mat)
    A_mat_sparse = scipy.sparse.csc_matrix(A_mat)

    m = osqp.OSQP()

    m.setup(
        P=P_mat_sparse,
        q=q_vec,
        A=A_mat_sparse,
        rho=0.1,
        sigma=5e-10,
        l=l_vec,
        u=u_vec,
        eps_abs=eps_abs,
        eps_rel=eps_rel,
        delta=1e-07,
        # polish=True,
        polish=False,
        adaptive_rho=False,
        warm_start=True,
        verbose=False,
        max_iter=max_iter,
    )

    solve_res = m.solve()

    if solve_res.info.status_val == -2 and verbose:
        print("ERROR! OSQP Solver hit max iteration limit. Either reduce your tolerances or increase the max iterations!")

    return (solve_res, var_to_index_dict)


def update_osqp_vars(var_to_osqp_indices_dict, solver_values):
    """
    Updates the variables values based on the OSQP solution
    """
    for osqp_var in var_to_osqp_indices_dict.keys():
        osqp_var.val = solver_values[var_to_osqp_indices_dict[osqp_var]]

def print_osqp_vars_and_sol(solve_res_x, var_to_index_dict):
    for key, idx in var_to_index_dict.items():
        print(f"{key}, {solve_res_x[idx]}")



import numdifftools as nd
import numpy as np
from scipy.linalg import eigvalsh

DEFAULT_TOL = 1e-4

"""
Utility classes to represent expresions. Each expression defines an eval, grad,
hess, and convexify method. Variables and values are assumed to be 2D numpy
arrays.
"""

N_DIGS = 6  # 10


class Expr(object):

    """
    by default, expressions are defined by black box functions
    """

    def __init__(self, f, grad=None, hess=None, **kwargs):
        self.f = f
        self._grad = grad
        self._hess = hess

        self._eval_cache = {}
        self._grad_cache = {}
        self._convexify_cache = {}

    def _get_key(self, x):
        return tuple(x.round(N_DIGS).flatten())

    def eval(self, x):
        key = self._get_key(x)

        if key in self._eval_cache:
            return self._eval_cache[key]
        val = self.f(x)
        self._eval_cache[key] = val.copy()
        return val

    def _get_flat_f(self, x):
        """
        Utility function which reshapes and flattens for compatibility with
        numdifftools' Jacobian and Hessian function.
        """
        if len(x.shape) == 2:
            rows, cols = x.shape
            assert cols == 1

            def flat_f(x):
                return self.f(x.reshape((rows, cols))).flatten()

            return flat_f
        elif len(x.shape) == 1:
            return self.f
        else:
            raise Exception("Input shape not supported")

    def _num_grad(self, x):
        """
        Returns a numerically computed gradient.
        Flattening is necessary for compatibility with numdifftools' Jacobian
        function.
        """
        grad_fn = nd.Jacobian(self._get_flat_f(x))

        return grad_fn(x)

    def _debug_grad(self, g1, g2, atol=DEFAULT_TOL):
        for i, g_row in enumerate(g1):
            for j, g in enumerate(g_row):
                if not np.allclose(g, g2[i, j], atol=atol):
                    print("{}, {}".format(i, j))
                    print(g, g2[i, j])

    def grad(self, x, num_check=False, atol=DEFAULT_TOL):
        """
        Returns the gradient. Can numerically check the gradient.
        """
        key = self._get_key(x)
        if key in self._grad_cache:
            return self._grad_cache[key].copy()
        assert not num_check or self._grad is not None
        if self._grad is None:
            return self._num_grad(x)
        gradient = self._grad(x)
        if num_check:
            num_grad = self._num_grad(x)
            if not np.allclose(num_grad, gradient, atol=atol):
                self._debug_grad(gradient, num_grad, atol=atol)
                raise Exception(
                    "Numerical and analytical gradients aren't close. \
                    \nnum_grad: {0}\nana_grad: {1}\n".format(
                        num_grad, gradient
                    )
                )
        self._grad_cache[key] = gradient.copy()
        return gradient

    def _num_hess(self, x):
        """
        Returns a numerically computed hessian.
        Flattening is necessary for compatibility with numdifftools' Hessian
        function.
        """
        hess_fn = nd.Hessian(self._get_flat_f(x))
        return hess_fn(x.flatten())

    def hess(self, x, num_check=False, atol=DEFAULT_TOL):
        """
        Returns the hessian. Can numerically check the hessian.
        """
        assert not num_check or self._hess is not None
        if self._hess is None:
            return self._num_hess(x)
        hessian = self._hess(x)
        if num_check:
            num_hess = self._num_hess(x)
            if not np.allclose(num_hess, hessian, atol=atol):
                raise Exception(
                    "Numerical and analytical hessians aren't close. \
                    \nnum_hess: {0}\nana_hess: {1}\n".format(
                        num_hess, hessian
                    )
                )
        return hessian

    def convexify(self, x, degree=1):
        """
        Returns an Expression object that represents the convex approximation of
        self at x where degree 1 is an affine approximation and degree 2 is a
        quadratic approximation. If the hessian has negative eigenvalues, the
        hessian is adjusted so that it is positive semi-definite.
        """

        res = None
        if degree == 1:
            A = self.grad(x)
            b = -A.dot(x) + self.eval(x)
            res = AffExpr(A, b)
        elif degree == 2:
            hess = self.hess(x)
            eig_vals = eigvalsh(hess)
            min_eig_val = min(eig_vals)
            if min_eig_val < 0:
                hess = hess - np.eye(hess.shape[0]) * min_eig_val
            grad = self.grad(x)
            Q = hess
            A = grad - np.transpose(x).dot(hess)
            b = 0.5 * np.transpose(x).dot(hess).dot(x) - grad.dot(x) + self.eval(x)
            res = QuadExpr(Q, A, b)
        else:
            raise NotImplementedError
        return res


class AffExpr(Expr):
    """
    Affine Expression
    """

    def __init__(self, A, b):
        """
        expr is Ax + b
        """
        assert b.shape[0] == A.shape[0]

        self.A = A
        self.b = b
        self.x_shape = (A.shape[1], 1)

    def eval(self, x):
        return self.A.dot(x) + self.b

    def grad(self, x):
        return self.A.T

    def hess(self, x):
        return np.zeros((self.x_shape[0], self.x_shape[0]))


class QuadExpr(Expr):
    """
    Quadratic Expression
    """

    def __init__(self, Q, A, b):
        """
        expr is 0.5*x'Qx + Ax + b
        """
        assert A.shape[0] == 1, "Can only define scalar quadrative expressions"

        # ensure the correct shapes for all the arguments
        assert Q.shape[0] == Q.shape[1]
        assert Q.shape[0] == A.shape[1]
        assert b.shape[0] == 1

        self.Q = Q
        self.A = A
        self.b = b
        self.x_shape = (A.shape[1], 1)

    def eval(self, x):
        return 0.5 * x.T.dot(self.Q.dot(x)) + self.A.dot(x) + self.b

    def grad(self, x):
        assert x.shape == self.x_shape
        return 0.5 * (self.Q.dot(x) + self.Q.T.dot(x)) + self.A.T

    def hess(self, x):
        return self.Q.copy()


class AbsExpr(Expr):
    """
    Absolute value expression
    """

    def __init__(self, expr):
        self.expr = expr

    def eval(self, x):
        return np.absolute(self.expr.eval(x))

    def grad(self, x):
        """
        Since the absolute value expression is not smooth, a subgradient is
        returned instead of the gradient.
        """
        raise NotImplementedError

    def hess(self, x):
        raise NotImplementedError


class HingeExpr(Expr):
    """
    Hinge expression
    """

    def __init__(self, expr):
        self.expr = expr

    def eval(self, x):
        v = self.expr.eval(x)
        zeros = np.zeros(v.shape)
        return np.maximum(v, zeros)

    def grad(self, x):
        """
        Since the hinge expression is not smooth, a subgradient is returned
        instead of the gradient.
        """
        raise NotImplementedError

    def hess(self, x):
        raise NotImplementedError


class CompExpr(Expr):
    """
    Comparison Expression
    """

    def __init__(self, expr, val):
        """
        expr: Expr object, the expression that is being compared to val
        val: numpy array, the value that the expression is being compared to
        """
        self.expr = expr
        self.val = val.copy()
        self._convexify_cache = {}

    def eval(self, x, tol=DEFAULT_TOL):
        """
        Returns True if the comparison holds true within some tolerace and 0
        otherwise.
        """
        raise NotImplementedError

    def grad(self, x):
        raise Exception(
            "The gradient is not well defined for comparison \
            expressions"
        )

    def hess(self, x):
        raise Exception(
            "The hessian is not well defined for comparison \
            expressions"
        )

    def convexify(self, x, degree=1):
        raise NotImplementedError


class EqExpr(CompExpr):
    """
    Equality Expression
    """

    def eval(self, x, tol=DEFAULT_TOL, negated=False):
        """
        Tests whether the expression at x is equal to self.val with tolerance
        tol.
        """
        assert tol >= 0.0
        if negated:
            return not np.allclose(self.expr.eval(x), self.val, atol=tol)
        return np.allclose(self.expr.eval(x), self.val, atol=tol)

    def convexify(self, x, degree=1):
        """
        Returns an AbsExpr that is the l1 penalty expression, a measure of
        constraint violation.

        The constraint h(x) = 0 becomes |h(x)|
        """
        assert degree == 1

        key = self._get_key(x)
        if key in self._convexify_cache:
            return self._convexify_cache[key]

        aff_expr = self.expr.convexify(x, degree=1)
        aff_expr.b = aff_expr.b - self.val
        res = AbsExpr(aff_expr)

        self._convexify_cache[key] = res
        return res


class LEqExpr(CompExpr):
    """
    Less than or equal to expression
    """

    def eval(self, x, tol=DEFAULT_TOL, negated=False):
        """
        Tests whether the expression at x is less than or equal to self.val with
        tolerance tol.
        """
        assert tol >= 0.0
        expr_val = self.expr.eval(x)
        if negated:
            ## need the tolerance to go the other way if its negated
            return not np.all(expr_val <= self.val - tol * np.ones(expr_val.shape))
        else:
            return np.all(expr_val <= self.val + tol * np.ones(expr_val.shape))

    def convexify(self, x, degree=1):
        """
        Returns a HingeExpr that is the hinge penalty expression, a measure of
        constraint violation.

        The constraint g(x) <= 0 becomes |g(x)|+ where |g(x)|+ = max(g(x), 0)
        """
        assert degree == 1

        key = self._get_key(x)
        if key in self._convexify_cache:
            return self._convexify_cache[key]

        aff_expr = self.expr.convexify(x, degree=1)
        aff_expr.b = aff_expr.b - self.val
        res = HingeExpr(aff_expr)

        self._convexify_cache[key] = res
        return res


class LExpr(CompExpr):
    """
    Less than expression
    """

    def eval(self, x, tol=DEFAULT_TOL, negated=False):
        """
        Tests whether the expression at x is less than or equal to self.val with
        tolerance tol.
        """
        assert tol >= 0.0
        expr_val = self.expr.eval(x)
        if negated:
            ## need the tolerance to go the other way if its negated
            return not np.all(expr_val < self.val - tol * np.ones(expr_val.shape))
        else:
            return np.all(expr_val < self.val + tol * np.ones(expr_val.shape))

    def convexify(self, x, degree=1):
        """
        Returns a HingeExpr that is the hinge penalty expression, a measure of
        constraint violation.

        The constraint g(x) <= 0 becomes |g(x)|+ where |g(x)|+ = max(g(x), 0)
        """
        assert degree == 1

        key = self._get_key(x)
        if key in self._convexify_cache:
            return self._convexify_cache[key]

        aff_expr = self.expr.convexify(x, degree=1)
        aff_expr.b = aff_expr.b - self.val
        res = HingeExpr(aff_expr)

        self._convexify_cache[key] = res
        return res


class BoundExpr(object):
    """
    Bound expression

    Bound expression is composed of an Expr and a Variable. Please note that the
    variable ordering matters
    """

    def __init__(self, expr, var):
        self.expr = expr
        self.var = var

    def eval(self):
        """
        Returns the current value of the bound expression
        """
        return self.expr.eval(self.var.get_value())

    def convexify(self, degree=1):
        """
        Returns a convexified BoundExpr at the variable's current value.
        """
        assert self.var.get_value() is not None
        cvx_expr = self.expr.convexify(self.var.get_value(), degree)
        return BoundExpr(cvx_expr, self.var)


class TFExpr(Expr):

    """
    TODO

    wrapper around exprs defined by a tensorflow graph. Leverages
    automated differentition.
    """

    def __init__(self, f, grad=None, hess=None, sess=None):
        self.sess = sess
        return super(TFExpr, self).__init__(f, grad, hess)



# fmt: off
from collections import defaultdict

import gurobipy as grb
import numpy as np

from sco_py.expr import (AbsExpr, AffExpr, CompExpr, EqExpr, HingeExpr,
                             LEqExpr, QuadExpr)

# fmt: on

GRB = grb.GRB

class Prob(object):
    """
    Sequential convex programming problem with a scalar objective. A solution is
    found using the l1 penalty method.
    """

    def __init__(self, grb_model, callback=None):
        """
        _model: Gurobi model associated with this problem
        _vars: variables in this problem

        _quad_obj_exprs: list of quadratic expressions in the objective
        _nonquad_obj_exprs: list of non-quadratic expressions in the objective
        _approx_obj_exprs: list of approximation of the non-quadratic
            expressions in the objective

        _nonlin_cnt_exprs: list of non-linear constraint expressions
        _penalty_exprs: list of penalty term expressions (approximations of the
            non-linear constraint expressions in _nonlin_cnt_exprs)

        _grb_penalty_cnts: list of Gurobi constraints that are generated when
            adding the hinge and absolute value terms from the penalty terms.
        _pgm: Positive Gurobi variable manager provides a lazy way of generating
            positive Gurobi variables so that there are less model updates.

        _bexpr_to_grb_expr: dictionary that caches quadratic bound expressions
            with their corresponding Gurobi expression
        """
        self._model = grb_model
        self._model.params.OutputFlag = 0  # silences Gurobi output
        self._vars = set()
        if callback is not None:
            self._callback = callback
        else:

            def do_nothing():
                pass

            self._callback = do_nothing

        self._quad_obj_exprs = []
        self._nonquad_obj_exprs = []
        self._approx_obj_exprs = []

        # linear constraints are added directly to the model so there's no
        # need for a _lin_cnt_exprs variable
        self._nonlin_cnt_exprs = []

        # list of constraints that will hold the hinge constraints
        # for each non-linear constraint, is a pair of constraints
        # for an eq constraint
        self.hinge_created = False

        self._penalty_exprs = []
        self._grb_penalty_cnts = []  # hinge and abs value constraints
        self._pgm = PosGRBVarManager(self._model)

        self._bexpr_to_grb_expr = {}

        ## group-id (str) -> cnt-set (set of constraints)
        self._cnt_groups = defaultdict(set)
        self._cnt_groups_overlap = defaultdict(set)
        self._penalty_groups = []
        self.nonconverged_groups = []
        self.gid2ind = {}

    def add_obj_expr(self, bound_expr):
        """
        Adds a bound expression (bound_expr) to the objective. If the objective
        is quadratic, is it added to _quad_obj_exprs. Otherwise, it is added
        to self._nonquad_obj_exprs.

        bound_expr's var is added to self._vars so that a trust region can be
        added to var.
        """
        expr = bound_expr.expr
        if isinstance(expr, AffExpr) or isinstance(expr, QuadExpr):
            self._quad_obj_exprs.append(bound_expr)
        else:
            self._nonquad_obj_exprs.append(bound_expr)
        self.add_var(bound_expr.var)

    def add_var(self, var):
        self._vars.add(var)

    def add_cnt_expr(self, bound_expr, group_ids=None):
        """
        Adds a bound expression (bound_expr) to the problem's constraints.
        If the constraint is linear, it is added directly to the model.
        Otherwise, the constraint is added by appending bound_expr to
        self._nonlin_cnt_exprs.

        bound_expr's var is added to self._vars so that a trust region can be
        added to var.
        """
        comp_expr = bound_expr.expr
        expr = comp_expr.expr
        var = bound_expr.var
        if isinstance(expr, AffExpr):
            # adding constraint directly into model
            grb_expr, grb_cnt = self._aff_expr_to_grb_expr(expr, var)
            if isinstance(comp_expr, EqExpr):
                self._add_np_array_grb_cnt(grb_expr, GRB.EQUAL, comp_expr.val)
            elif isinstance(comp_expr, LEqExpr):
                self._add_np_array_grb_cnt(grb_expr, GRB.LESS_EQUAL, comp_expr.val)
        else:
            self._nonlin_cnt_exprs.append(bound_expr)
            self._reset_hinge_cnts()

            if group_ids is None:
                group_ids = ["all"]
            for gid in group_ids:
                self._cnt_groups[gid].add(bound_expr)
                for other in group_ids:
                    if other == gid:
                        continue
                    self._cnt_groups_overlap[gid].add(other)

        self.add_var(var)

    def _reset_hinge_cnts(self):
        ## reset the hinge_cnts
        self.hinge_created = False

    # @profile
    def _add_np_array_grb_cnt(self, grb_exprs, sense, val):
        """
        Adds a numpy array of Gurobi constraints to the model and returns
        the constraints.
        """
        cnts = []
        for index, grb_expr in np.ndenumerate(grb_exprs):
            cnts.append(self._model.addConstr(grb_expr, sense, val[index]))
        return cnts

    # @profile
    def _expr_to_grb_expr(self, bound_expr):
        """
        Translates AffExpr, QuadExpr, HingeExpr and AbsExpr to Gurobi
        expressions and returns the corresponding Gurobi expressions and
        constraints. If there are no Gurobi constraints, an empty list is
        returned. Otherwise, this method raises an exception.
        """
        expr = bound_expr.expr
        var = bound_expr.var

        if isinstance(expr, AffExpr):
            return self._aff_expr_to_grb_expr(expr, var)
        elif isinstance(expr, QuadExpr):
            if bound_expr in self._bexpr_to_grb_expr:
                return self._bexpr_to_grb_expr[bound_expr], []
            else:
                grb_expr, cnts = self._quad_expr_to_grb_expr(expr, var)
                self._bexpr_to_grb_expr[bound_expr] = grb_expr
                return grb_expr, cnts
        elif isinstance(expr, HingeExpr):
            return self._hinge_expr_to_grb_expr(expr, var)
        elif isinstance(expr, AbsExpr):
            return self._abs_expr_to_grb_expr(expr, var)
        elif isinstance(expr, CompExpr):
            raise Exception(
                "Comparison Expressions cannot be converted to \
                a Gurobi expression. Use add_cnt_expr instead"
            )
        else:
            raise Exception(
                "This type of Expression cannot be converted to\
                a Gurobi expression."
            )

    # @profile
    def _aff_expr_to_grb_expr(self, aff_expr, var):
        grb_var = var.get_grb_vars()
        grb_exprs = []
        A = aff_expr.A
        b = aff_expr.b
        for i in range(A.shape[0]):
            grb_expr = grb.LinExpr()
            (inds,) = np.nonzero(A[i, :])
            grb_expr += b[i]
            grb_expr.addTerms(A[i, inds].tolist(), grb_var[inds, 0].tolist())
            grb_exprs.append([grb_expr])
        return np.array(grb_exprs), []

    # #@profile
    def _quad_expr_to_grb_expr(self, quad_expr, var):
        x = var.get_grb_vars()
        grb_expr = grb.QuadExpr()
        Q = quad_expr.Q
        rows, cols = x.shape
        assert cols == 1
        inds = np.nonzero(Q)
        coeffs = 0.5 * Q[inds]
        v1 = x[inds[0], 0]
        v2 = x[inds[1], 0]
        grb_expr.addTerms(coeffs.tolist(), v1.tolist(), v2.tolist())
        inds = np.nonzero(quad_expr.A)
        coeffs = quad_expr.A[inds]
        v1 = x[inds[1], 0]
        grb_expr.addTerms(coeffs.tolist(), v1.tolist())
        grb_expr = grb_expr + quad_expr.b

        return np.array([[grb_expr]]), []

    # @profile
    def _hinge_expr_to_grb_expr(self, hinge_expr, var):
        aff_expr = hinge_expr.expr
        assert isinstance(aff_expr, AffExpr)
        grb_expr, _ = self._aff_expr_to_grb_expr(aff_expr, var)
        grb_hinge = self._pgm.get_array(grb_expr.shape)
        cnts = self._add_np_array_grb_cnt(grb_expr, GRB.LESS_EQUAL, grb_hinge)

        return grb_hinge, cnts

    def _abs_expr_to_grb_expr(self, abs_expr, var):
        aff_expr = abs_expr.expr
        assert isinstance(aff_expr, AffExpr)
        grb_expr, _ = self._aff_expr_to_grb_expr(aff_expr, var)
        pos = self._pgm.get_array(grb_expr.shape)
        neg = self._pgm.get_array(grb_expr.shape)
        cnts = self._add_np_array_grb_cnt(grb_expr, GRB.EQUAL, pos - neg)

        return pos + neg, cnts

    def find_closest_feasible_point(self):
        """
        Finds the closest point (l2 norm) to the initialization that satisfies
        the linear constraints.
        """
        self._del_old_grb_cnts()
        self._model.update()

        obj = grb.QuadExpr()
        for var in self._vars:
            g_var = var.get_grb_vars()
            val = var.get_value()
            if val is not None:
                assert g_var.shape == val.shape
                inds = np.where(~np.isnan(val))
                val = val[inds]
                g_var = g_var[inds]
                obj += np.sum((g_var - val).T.dot(g_var - val))

                # for i in np.ndindex(g_var.shape):
                #    if not np.isnan(val[i]):
                #            obj += g_var[i]*g_var[i] - 2*val[i]*g_var[i] + val[i]*val[i]

            # grb_exprs = []
            # for bound_expr in self._quad_obj_exprs:
            #     grb_expr, grb_cnts = self._expr_to_grb_expr(bound_expr)
            #     self._grb_penalty_cnts.extend(grb_cnts)
            #     grb_exprs.extend(grb_expr.flatten().tolist())

            # obj += grb.quicksum(grb_exprs)

        self._model.setObjective(obj)

        self.optimize()

        return self._model.status == 2

    def optimize(self):
        """
        Calls the Gurobi optimizer on the current QP approximation with a given
        penalty coefficient.

        Temporary Gurobi constraints and variables from the previous optimize
        call are deleted.

        The Gurobi objective is computed by translating all the expressions
        in the quadratically approximated objective (self._quad_obj_expr) and
        in the penalty approximation of the constraints (self._penalty_exprs)
        to Gurobi expressions, and summing them. The temporary constraints and
        variables created from the translation process are saved so that they
        can be deleted later.

        The Gurobi constraints are the linear constraints which have already
        been added to the model when constraints were added to this problem.
        """
        self._model.optimize()
        try:
            self._update_vars()
        except Exception as e:
            print(e)
            print(("Model status:", self._model.status))
        self._callback()

    def print_grb_vals(self):
        """convenience function for debugging"""
        for var in self._model.getVars():
            print(var)

    # @profile
    def update_obj(self, penalty_coeff=0.0):
        self._lazy_spawn_grb_cnts()
        grb_exprs = []
        
        for bound_expr in self._quad_obj_exprs + self._approx_obj_exprs:
            grb_expr, grb_cnts = self._expr_to_grb_expr(bound_expr)
            # self._grb_penalty_cnts.extend(grb_cnts)
            grb_exprs.extend(grb_expr.flatten().tolist())

        for i, bound_expr in enumerate(self._penalty_exprs):
            grb_expr = self._update_nonlin_cnt(bound_expr, i).flatten()
            grb_exprs.extend(grb_expr * penalty_coeff)

        obj = grb.quicksum(grb_exprs)

        self._model.setObjective(obj)
        self._model.update()

    def _del_old_grb_cnts(self):
        for cnt in self._grb_penalty_cnts:
            self._model.remove(cnt)
        self.hinge_created = False

    def _lazy_spawn_grb_cnts(self):
        if self.hinge_created:
            return
        self._del_old_grb_cnts()
        self._grb_penalty_cnts = []
        self._grb_penalty_exprs = []
        self._grb_nz = []
        for bound_expr in self._penalty_exprs:
            grb_expr, grb_cnts = self._expr_to_grb_expr(bound_expr)
            self._grb_penalty_cnts.append(grb_cnts)
            self._grb_penalty_exprs.append(grb_expr)
            self._grb_nz.append(np.nonzero(bound_expr.expr.expr.A))
        self._model.update()
        self.hinge_created = True

    # @profile
    def _update_nonlin_cnt(self, bexpr, ind):
        expr, var = bexpr.expr, bexpr.var
        if isinstance(expr, HingeExpr) or isinstance(expr, AbsExpr):
            aff_expr = expr.expr
            assert isinstance(aff_expr, AffExpr)
            A, b = aff_expr.A, aff_expr.b
            cnts = self._grb_penalty_cnts[ind]
            grb_expr = self._grb_penalty_exprs[ind]
            old_nz = self._grb_nz[ind]
            grb_vars = var.get_grb_vars()
            nz = np.nonzero(A)

            # This updates the constraint's RHS
            for i in range(A.shape[0]):
                ## add negative b to rhs because it
                ## changes sides of the ineq/eq
                cnts[i].setAttr("rhs", -b[i, 0])

            # This sets all the old nz terms' coeffs to 0
            for idx in range(old_nz[0].shape[0]):
                i, j = old_nz[0][idx], old_nz[1][idx]
                self._model.chgCoeff(cnts[i], grb_vars[j, 0], 0)

            # The nonzero terms are then updated with their new coeffs
            ## then set the non-zero values
            for idx in range(nz[0].shape[0]):
                i, j = nz[0][idx], nz[1][idx]
                self._model.chgCoeff(cnts[i], grb_vars[j, 0], A[i, j])

            self._grb_nz[ind] = nz

            return grb_expr
        else:
            raise NotImplementedError

    def add_trust_region(self, trust_region_size):
        """
        Adds the trust region for every variable
        """
        for var in self._vars:
            var.add_trust_region(trust_region_size)

    # @profile
    def convexify(self):
        """
        Convexifies the optimization problem by computing a QP approximation
        A quadratic approximation of the non-quadratic objective terms
        (self._nonquad_obj_exprs) is saved in self._approx_obj_exprs.
        The penalty approximation of the non-linear constraints
        (self._nonlin_cnt_exprs) is saved in self._penalty_exprs
        """
        # if len(self._nonlin_cnt_exprs) > 0:
        #     import ipdb; ipdb.set_trace()
        self._approx_obj_exprs = [
            bexpr.convexify(degree=2) for bexpr in self._nonquad_obj_exprs
        ]
        self._penalty_exprs = [
            bexpr.convexify(degree=1) for bexpr in self._nonlin_cnt_exprs
        ]
        self._penalty_groups = []
        gids = sorted(self._cnt_groups.keys())
        self.gid2ind = {}
        for i, gid in enumerate(gids):
            self.gid2ind[gid] = i
            cur_bexprs = [bexpr.convexify(degree=1) for bexpr in self._cnt_groups[gid]]
            self._penalty_groups.append(cur_bexprs)

    # #@profile
    def get_value(self, penalty_coeff, vectorize=False):
        """
        Returns the current value of the penalty objective.
        The penalty objective is computed by summing up all the values of the
        quadratic objective expressions (self._quad_obj_exprs), the
        non-quadratic objective expressions and the penalty coeff multiplied
        by the constraint violations (computed using _nonlin_cnt_exprs)

        if vectorize=True, then this returns a vector of constraint
        violations -- 1 per group id.
        """
        if vectorize:
            gids = sorted(self._cnt_groups.keys())
            value = np.zeros(len(gids))
            for i, gid in enumerate(gids):
                value[i] = np.sum(
                    np.sum(
                        [
                            np.sum(self._compute_cnt_violation(bexpr))
                            for bexpr in self._cnt_groups[gid]
                        ]
                    )
                )
            return value
        else:
            value = 0.0
            for bound_expr in self._quad_obj_exprs + self._nonquad_obj_exprs:
                value += np.sum(np.sum(bound_expr.eval()))
            for bound_expr in self._nonlin_cnt_exprs:
                cnt_vio = self._compute_cnt_violation(bound_expr)
                value += penalty_coeff * np.sum(cnt_vio)

            return value

    # @profile
    def _compute_cnt_violation(self, bexpr):
        comp_expr = bexpr.expr
        var_val = bexpr.var.get_value()
        if isinstance(comp_expr, EqExpr):
            return np.absolute(comp_expr.expr.eval(var_val) - comp_expr.val)
        elif isinstance(comp_expr, LEqExpr):
            v = comp_expr.expr.eval(var_val) - comp_expr.val
            zeros = np.zeros(v.shape)
            return np.maximum(v, zeros)

    def get_max_cnt_violation(self):
        """
        Returns the the maximum amount a non-linear constraint is violated.
        Linear constraints are assumed to be satisfied because they are added
        directly to the model and QP solvers can deal with them.
        """
        max_vio = 0.0
        for bound_expr in self._nonlin_cnt_exprs:
            cnt_vio = self._compute_cnt_violation(bound_expr)
            cnt_max_vio = np.amax(cnt_vio)
            max_vio = np.maximum(max_vio, cnt_max_vio)
        return max_vio

    def get_approx_value(self, penalty_coeff, vectorize=False):
        """
        Returns the current value of the penalty QP approximation by summing
        up the expression values for the quadratic objective terms
        (_quad_obj_exprs), the quadratic approximation of the non-quadratic
        terms (_approx_obj_exprs) and the penalty terms (_penalty_exprs).
        Note that this approximate value is computed with respect to when the
        last convexification was performed.

        if vectorize=True, then this returns a vector of constraint
        violations -- 1 per group id.
        """
        if vectorize:
            value = np.zeros(len(self._penalty_groups))
            for i, bexprs in enumerate(self._penalty_groups):
                x = np.array([np.sum(bexpr.eval()) for bexpr in bexprs])
                value[i] = np.sum(x.flatten())
            return value

        value = 0.0
        for bound_expr in self._quad_obj_exprs + self._approx_obj_exprs:
            value += np.sum(np.sum(bound_expr.eval()))
        for bound_expr in self._penalty_exprs:
            value += penalty_coeff * np.sum(bound_expr.eval())

        return value

    def _update_vars(self):
        """
        Updates the variables values
        """
        unique_grb_vars = set()
        for var in self._vars:
            var.update()
            for gv in var._grb_vars.flatten().tolist():
                unique_grb_vars.add(gv)

    def save(self):
        """
        Saves the problem's current state by saving the values of all the
        variables.
        """
        for var in self._vars:
            var.save()

    def restore(self):
        """
        Restores the problem's state to the problem's saved state
        """
        for var in self._vars:
            var.restore()


class PosGRBVarManager(object):
    """
    Manages positive Gurobi variables. The purpose of the manager is to create
    many Gurobi variables at once to decrease the number of Gurobi model update
    because model updates take a long time.
    """

    INIT_NUM = 1000
    INC_NUM = 1000

    def __init__(self, model, init_num=INC_NUM, inc_num=INC_NUM):
        self._index = 0
        self._model = model
        self._grb_vars = []
        self._add_grb_vars(init_num)
        self._inc_num = inc_num

    def _add_grb_vars(self, num=None):
        """
        Creates a batch of positive Gurobi variables so that the model is
        updated less often.
        """
        if num is None:
            num = self._inc_num
        new_grb_vars = [self._model.addVar(lb=0.0, ub=GRB.INFINITY) for i in range(num)]
        self._grb_vars.extend(new_grb_vars)
        self._model.update()

    def __next__(self):
        """
        Returns one positive Gurobi variable.
        """
        if self._index == len(self._grb_vars) - 1:
            self._add_grb_vars()
        self._index += 1
        return self._grb_vars[self._index - 1]

    # @profile
    def get_array(self, shape):
        """
        Returns a numpy array of unused positive Gurobi variables.
        """
        a = np.empty(shape, dtype=object)
        for x in np.nditer(a, op_flags=["readwrite"], flags=["refs_ok"]):
            x[...] = next(self)
        return a

    def reset(self):
        self._index = 0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sco_py', 'sco_py.sco_gurobi', 'sco_py.sco_osqp']

package_data = \
{'': ['*']}

install_requires = \
['gurobipy>=9.0.0,<9.2.0',
 'numdifftools>=0.9.40,<0.10.0',
 'numpy>=1.21.4,<2.0.0',
 'osqp>=0.6.2,<0.7.0',
 'pytest>=6.2.5,<7.0.0',
 'scipy>=1.7.2,<2.0.0']

setup_kwargs = {
    'name': 'sco-py',
    'version': '0.3.0',
    'description': 'A Sequential Convex Optimization library for solving non-convex optimization problems. Intended for use with the OpenTAMP planning system.',
    'long_description': "[![continuous-integration](https://github.com/Algorithmic-Alignment-Lab/sco_py/actions/workflows/ci.yaml/badge.svg)](https://github.com/Algorithmic-Alignment-Lab/sco_py/actions/workflows/ci.yaml) [![PyPI version](https://badge.fury.io/py/sco-py.svg)](https://badge.fury.io/py/sco-py)\n\n# sco_py: Sequential Convex Optimization with Python\nsco_py is a lightweight Sequential Convex Optimization library for solving non-convex optimization problems. sco_py is intended for use with the OpenTAMP planning system. Currently, the library supports both [Gurobi](https://www.gurobi.com/) (license required) and [OSQP](https://osqp.org/) (open-source, no license required!) as backend QP solvers.\n\n## Installation\n### From PyPI with pip\nSimply run: `pip install sco-py`\n\n### From GitHub with pip\nSimply run: `pip install git+https://github.com/Algorithmic-Alignment-Lab/sco_py.git`\n\n### Developer (from source)\n1. Clone this repository [from GitHub](https://github.com/Algorithmic-Alignment-Lab/sco_py)\n1. Install Poetry by following the instructions from [here](https://python-poetry.org/docs/#installation)\n1. Install all dependencies with `poetry install`.\n\n## Contributing\nsco is an open-source repository and as such, we welcome contributions from interested members of the community! If you have a new idea for a feature/contribution, do post in the ['Discussions' tab of the GitHub repository](https://github.com/Algorithmic-Alignment-Lab/sco_py/discussions) to get some feedback from the maintainers before starting development. In general, we recommend that you fork the main repository, create a new branch with your proposed change, then open a pull-request into the main repository. The main requirement for a new feature is that it cannot break the current test cases (see below for how to run our tests) unless this is unavoidable, and in this case, it should modify/introduce new tests as necessary. In particular, we welcome updates to documentation (docstrings, comments, etc. that make the codem more approachable for new users) and test cases!\n\n### Running tests\nIf you do not have a license for Gurobi, then you can only run the OSQP tests. To do so, run:\n```\npytest tests/sco_osqp/\n```\nIf you do have a license for Gurobi, then you can run all tests with the `pytest` command.\n\nNote that our Contrinuous Integration (CI) setup only checks and reports status for OSQP tests. In general, if you are contributing a new feature, it *must* pass the existing OSQP tests and contribute new tests that test the new feature at least with OSQP (and preferably with Gurobi as well).",
    'author': 'Nishanth Kumar',
    'author_email': 'njk@csail.mit.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



import numpy as np


class Variable(object):
    """
    Variable

    Manages Gurobi variables by maintaining an ordering of Gurobi variables,
    """

    def __init__(self, grb_vars, value=None):
        """
        _grb_vars: Numpy array of Gurobi variables. The ordering of the Gurobi
        variables must be maintained.
        _value: Numpy array of the current value of this variable
        _saved_value: saved value of this variable
        """
        assert isinstance(grb_vars, np.ndarray)
        assert len(grb_vars) > 0
        self._grb_vars = grb_vars.copy()
        if value is not None:
            assert grb_vars.shape == value.shape
            assert isinstance(value, np.ndarray)
            self._value = value.copy()
        else:
            self._value = None
        self._saved_value = None

    def get_grb_vars(self):
        return self._grb_vars.copy()

    def get_value(self):
        if self._value is not None:
            return self._value.copy()
        else:
            return None

    def add_trust_region(self, trust_box_size):
        """
        Adds a trust region around the saved value (self._saved_value) by
        changing the upper and lower bounds of the Gurobi variables
        self._grb_vars
        """
        assert self._saved_value is not None
        # assert not np.any(np.isnan(self._saved_value))
        for index, grb_var in np.ndenumerate(self._grb_vars):
            grb_var.lb = self._saved_value[index] - trust_box_size
            grb_var.ub = self._saved_value[index] + trust_box_size

    def update(self):
        """
        If the gurobi variables have valid values, update self._value to reflect
        the values in the gurobi variables.

        When the gurobi variables do not have valid values, update will raise a
        GurobiError
        """
        value = np.zeros(self._grb_vars.shape)
        for index, grb_var in np.ndenumerate(self._grb_vars):
            value[index] = grb_var.X
        self._value = value

    def save(self):
        """
        Save the current value.
        """
        # assert not np.any(np.isnan(self._value.copy()))
        self._saved_value = self._value.copy()

    def restore(self):
        """
        Restore value to the saved value.
        """
        self._value = self._saved_value.copy()
